@@ -32,8 +32,12 @@ from matplotlib.backends.backend_gtk3cairo import FigureCanvasGTK3Cairo as Figur
 from matplotlib.figure import Figure
 
 from pitivi.configure import get_pixmap_dir
+from pitivi.effects import ALLOWED_ONLY_ONCE_EFFECTS
+from pitivi.effects import AUDIO_EFFECT
+from pitivi.effects import VIDEO_EFFECT
 from pitivi.timeline.previewers import AudioPreviewer
 from pitivi.timeline.previewers import VideoPreviewer
+from pitivi.undo.timeline import CommitTimelineFinalizingAction
 from pitivi.utils.loggable import Loggable
 from pitivi.utils.misc import disconnectAllByFunc
 from pitivi.utils.misc import filename_from_uri
@@ -107,38 +111,38 @@ class KeyframeCurve(FigureCanvas, Loggable):
         self.__line_xs = []
         self.__line_ys = []
 
-        # axisbg to None for transparency
-        self.__ax = figure.add_axes([0, 0, 1, 1], axisbg='None')
+        # facecolor to None for transparency
+        self._ax = figure.add_axes([0, 0, 1, 1], facecolor='None')
         # Clear the Axes object.
-        self.__ax.cla()
+        self._ax.cla()
 
         # FIXME: drawing a grid and ticks would be nice, but
         # matplotlib is too slow for now.
-        self.__ax.grid(False)
+        self._ax.grid(False)
 
-        self.__ax.tick_params(axis='both',
-                              which='both',
-                              bottom='off',
-                              top='off',
-                              right='off',
-                              left='off')
+        self._ax.tick_params(axis='both',
+                             which='both',
+                             bottom='off',
+                             top='off',
+                             right='off',
+                             left='off')
 
         # This seems to also be necessary for transparency ..
         figure.patch.set_visible(False)
 
         # The PathCollection object holding the keyframes dots.
         sizes = [50]
-        self.__keyframes = self.__ax.scatter([], [], marker='D', s=sizes,
-                                             c=KEYFRAME_NODE_COLOR, zorder=2)
+        self.__keyframes = self._ax.scatter([], [], marker='D', s=sizes,
+                                            c=KEYFRAME_NODE_COLOR, zorder=2)
 
         # matplotlib weirdness, simply here to avoid a warning ..
         self.__keyframes.set_picker(True)
 
         # The Line2D object holding the lines between keyframes.
-        self.__line = self.__ax.plot([], [],
-                                     alpha=KEYFRAME_LINE_ALPHA,
-                                     c=KEYFRAME_LINE_COLOR,
-                                     linewidth=KEYFRAME_LINE_HEIGHT, zorder=1)[0]
+        self.__line = self._ax.plot([], [],
+                                    alpha=KEYFRAME_LINE_ALPHA,
+                                    c=KEYFRAME_LINE_COLOR,
+                                    linewidth=KEYFRAME_LINE_HEIGHT, zorder=1)[0]
         self.__updatePlots()
 
         # Drag and drop logic
@@ -153,16 +157,13 @@ class KeyframeCurve(FigureCanvas, Loggable):
 
         self.__hovered = False
 
-        # Whether a keyframe has just been removed.
-        self.__keyframe_removed = False
-
         self.connect("motion-notify-event", self.__gtkMotionEventCb)
         self.connect("event", self._eventCb)
         self.connect("notify::height-request", self.__heightRequestCb)
 
-        self.mpl_connect('button_press_event', self.__mplButtonPressEventCb)
-        self.mpl_connect('button_release_event', self.__mplButtonReleaseEventCb)
-        self.mpl_connect('motion_notify_event', self.__mplMotionEventCb)
+        self.mpl_connect('button_press_event', self._mpl_button_press_event_cb)
+        self.mpl_connect('button_release_event', self._mpl_button_release_event_cb)
+        self.mpl_connect('motion_notify_event', self._mpl_motion_event_cb)
 
     def release(self):
         disconnectAllByFunc(self, self.__heightRequestCb)
@@ -178,7 +179,7 @@ class KeyframeCurve(FigureCanvas, Loggable):
 
         ylim_min = -(KEYFRAME_LINE_HEIGHT / height)
         ylim_max = (self.__ylim_max * height) / (height - KEYFRAME_LINE_HEIGHT)
-        self.__ax.set_ylim(ylim_min, ylim_max)
+        self._ax.set_ylim(ylim_min, ylim_max)
 
     def __heightRequestCb(self, unused_self, unused_pspec):
         self.__computeYlim()
@@ -195,7 +196,7 @@ class KeyframeCurve(FigureCanvas, Loggable):
             self.__line_xs.append(value.timestamp)
             self.__line_ys.append(value.value)
 
-        self.__ax.set_xlim(self.__line_xs[0], self.__line_xs[-1])
+        self._ax.set_xlim(self.__line_xs[0], self.__line_xs[-1])
         self.__computeYlim()
 
         arr = numpy.array((self.__line_xs, self.__line_ys))
@@ -212,7 +213,8 @@ class KeyframeCurve(FigureCanvas, Loggable):
             res, value = self.__source.control_source_get_value(event.xdata)
             assert res
             self.debug("Create keyframe at (%lf, %lf)", event.xdata, value)
-            with self.__timeline.app.action_log.started("Keyframe added"):
+            with self.__timeline.app.action_log.started("Keyframe added",
+                                                        toplevel=True):
                 self.__source.set(event.xdata, value)
 
     def toggle_keyframe(self, offset):
@@ -246,7 +248,7 @@ class KeyframeCurve(FigureCanvas, Loggable):
             self.__timeline.get_window().set_cursor(cursor)
         return False
 
-    def __mplButtonPressEventCb(self, event):
+    def _mpl_button_press_event_cb(self, event):
         if event.button != 1:
             return
 
@@ -262,14 +264,22 @@ class KeyframeCurve(FigureCanvas, Loggable):
                 if index == 0 or index == len(offsets) - 1:
                     # It's an edge keyframe. These should not be removed.
                     return
+
+                # Rollback the last operation if it is "Move keyframe".
+                # This is needed because a double-click also triggers a
+                # BUTTON_PRESS event which starts a "Move keyframe" operation
+                self.__timeline.app.action_log.try_rollback("Move keyframe")
+                self.__offset = None
+
                 # A keyframe has been double-clicked, remove it.
                 self.debug("Removing keyframe at timestamp %lf", offset)
-                self.__keyframe_removed = True
-                with self.__timeline.app.action_log.started("Remove keyframe"):
+                with self.__timeline.app.action_log.started("Remove keyframe",
+                                                            toplevel=True):
                     self.__source.unset(offset)
             else:
                 # Remember the clicked frame for drag&drop.
-                self.__timeline.app.action_log.begin("Move keyframe")
+                self.__timeline.app.action_log.begin("Move keyframe",
+                                                     toplevel=True)
                 self.__offset = offset
                 self.handling_motion = True
             return
@@ -278,7 +288,8 @@ class KeyframeCurve(FigureCanvas, Loggable):
         if result[0]:
             # The line has been clicked.
             self.debug("The keyframe curve has been clicked")
-            self.__timeline.app.action_log.begin("Move keyframe curve segment")
+            self.__timeline.app.action_log.begin("Move keyframe curve segment",
+                                                 toplevel=True)
             x = event.xdata
             offsets = self.__keyframes.get_offsets()
             keyframes = offsets[:, 0]
@@ -288,7 +299,7 @@ class KeyframeCurve(FigureCanvas, Loggable):
             self.__ydata_drag_start = max(self.__ylim_min, min(event.ydata, self.__ylim_max))
             self.handling_motion = True
 
-    def __mplMotionEventCb(self, event):
+    def _mpl_motion_event_cb(self, event):
         if event.ydata is not None and event.xdata is not None:
             # The mouse event is in the figure boundaries.
             if self.__offset is not None:
@@ -329,7 +340,7 @@ class KeyframeCurve(FigureCanvas, Loggable):
 
         self.__timeline.get_window().set_cursor(cursor)
 
-    def __mplButtonReleaseEventCb(self, event):
+    def _mpl_button_release_event_cb(self, event):
         if event.button != 1:
             return
 
@@ -340,19 +351,15 @@ class KeyframeCurve(FigureCanvas, Loggable):
             self.debug("Line released")
             self.__timeline.app.action_log.commit("Move keyframe curve segment")
 
+            if not self.__dragged:
+                # The keyframe line was clicked, but not dragged
+                assert event.guiEvent.type == Gdk.EventType.BUTTON_RELEASE
+                self.__maybeCreateKeyframe(event)
+
         self.handling_motion = False
         self.__offset = None
         self.__clicked_line = ()
-
-        if self.__dragged:
-            # The keyframe or keyframe line has already been dragged.
-            self.__dragged = False
-        else:
-            assert event.guiEvent.type == Gdk.EventType.BUTTON_RELEASE
-            if not self.__keyframe_removed:
-                self.__maybeCreateKeyframe(event)
-            else:
-                self.__keyframe_removed = False
+        self.__dragged = False
 
     def __update_tooltip(self, event):
         """Sets or clears the tooltip showing info about the hovered line."""
@@ -360,7 +367,10 @@ class KeyframeCurve(FigureCanvas, Loggable):
         if event:
             if not event.xdata:
                 return
-            xdata = max(self.__line_xs[0], min(event.xdata, self.__line_xs[-1]))
+            if self.__offset is not None:
+                xdata = self.__offset
+            else:
+                xdata = max(self.__line_xs[0], min(event.xdata, self.__line_xs[-1]))
             res, value = self.__source.control_source_get_value(xdata)
             assert res
             pmin = self.__paramspec.minimum
@@ -389,12 +399,12 @@ class KeyframeCurve(FigureCanvas, Loggable):
                 return event.xdata
 
             i = values.index(kf)
-            if event.xdata > self.__offset:
-                if values[i + 1].timestamp < event.xdata:
-                    return max(0, values[i + 1].timestamp - 1)
-            else:
-                if i > 1 and values[i - 1].timestamp > event.xdata:
-                    return values[i - 1].timestamp + 1
+            keyframe_timestamp = int(event.xdata)
+            if keyframe_timestamp <= values[i - 1].timestamp:
+                keyframe_timestamp = values[i - 1].timestamp + 1
+            if keyframe_timestamp >= values[i + 1].timestamp:
+                keyframe_timestamp = values[i + 1].timestamp - 1
+            return keyframe_timestamp
 
         return event.xdata
 
@@ -442,7 +452,7 @@ class TimelineElement(Gtk.Layout, Zoomable, Loggable):
         # We set up the default mixing property right here, if a binding was
         # already set (when loading a project), it will be added later
         # and override that one.
-        self.showDefaultKeyframes()
+        self.showDefaultKeyframes(lazy_render=True)
 
     def release(self):
         if self.__previewer:
@@ -467,9 +477,12 @@ class TimelineElement(Gtk.Layout, Zoomable, Loggable):
 
     def showKeyframes(self, ges_elem, prop):
         self.__setKeyframes(ges_elem, prop)
+        self.__create_keyframe_curve(ges_elem)
 
-    def showDefaultKeyframes(self):
+    def showDefaultKeyframes(self, lazy_render=False):
         self.__setKeyframes(self._ges_elem, self._getDefaultMixingProperty())
+        if not lazy_render:
+            self.__create_keyframe_curve()
 
     def __setKeyframes(self, ges_elem, prop):
         self.__removeKeyframes()
@@ -496,7 +509,8 @@ class TimelineElement(Gtk.Layout, Zoomable, Loggable):
         self.keyframe_curve = None
 
     # Private methods
-    def __createKeyframeCurve(self, binding):
+    def __ensure_keyframes(self, binding):
+        """Ensures that we have at least 2 keyframes (at inpoint and end)."""
         source = binding.props.control_source
         values = source.get_all()
 
@@ -505,12 +519,17 @@ class TimelineElement(Gtk.Layout, Zoomable, Loggable):
             val = float(self.__controlledProperty.default_value) / \
                 (self.__controlledProperty.maximum -
                  self.__controlledProperty.minimum)
-            source.set(self._ges_elem.props.in_point, val)
-            source.set(
-                self._ges_elem.props.duration + self._ges_elem.props.in_point,
-                val)
+            inpoint = self._ges_elem.props.in_point
+            assert source.set(inpoint, val)
+            assert source.set(inpoint + self._ges_elem.props.duration, val)
 
+    def __create_keyframe_curve(self, ges_elem=None):
+        """Creates required keyframe curve."""
         self.__removeKeyframes()
+        if not ges_elem:
+            ges_elem = self._ges_elem
+
+        binding = ges_elem.get_control_binding(self.__controlledProperty.name)
         self.keyframe_curve = KeyframeCurve(self.timeline, binding)
         self.keyframe_curve.connect("enter", self.__curveEnterCb)
         self.keyframe_curve.connect("leave", self.__curveLeaveCb)
@@ -519,6 +538,7 @@ class TimelineElement(Gtk.Layout, Zoomable, Loggable):
         self.__update_keyframe_curve_visibility()
 
     def __createControlBinding(self, element):
+        """Creates the required ControlBinding and keyframes."""
         if self.__controlledProperty:
             element.connect("control-binding-added",
                             self.__controlBindingAddedCb)
@@ -526,7 +546,7 @@ class TimelineElement(Gtk.Layout, Zoomable, Loggable):
                 element.get_control_binding(self.__controlledProperty.name)
 
             if binding:
-                self.__createKeyframeCurve(binding)
+                self.__ensure_keyframes(binding)
 
                 return
 
@@ -537,7 +557,7 @@ class TimelineElement(Gtk.Layout, Zoomable, Loggable):
 
     def __controlBindingAddedCb(self, unused_ges_elem, binding):
         if binding.props.name == self.__controlledProperty.name:
-            self.__createKeyframeCurve(binding)
+            self.__ensure_keyframes(binding)
 
     def do_draw(self, cr):
         self.propagate_draw(self.__background, cr)
@@ -552,6 +572,9 @@ class TimelineElement(Gtk.Layout, Zoomable, Loggable):
 
     # Callbacks
     def __selectedChangedCb(self, unused_selected, selected):
+        if not self.keyframe_curve and self.__controlledProperty:
+            self.__create_keyframe_curve()
+
         if self.keyframe_curve:
             self.__update_keyframe_curve_visibility()
 
@@ -561,7 +584,8 @@ class TimelineElement(Gtk.Layout, Zoomable, Loggable):
     def __update_keyframe_curve_visibility(self):
         """Updates the keyframes widget visibility by adding or removing it."""
         if self._ges_elem.selected and len(self.timeline.selection) == 1:
-            self.add(self.keyframe_curve)
+            if not self.keyframe_curve.get_parent():
+                self.add(self.keyframe_curve)
         else:
             self.remove(self.keyframe_curve)
 
@@ -755,7 +779,7 @@ class VideoUriSource(VideoSource):
         self.get_style_context().add_class("VideoUriSource")
 
     def _getPreviewer(self):
-        previewer = VideoPreviewer(self._ges_elem)
+        previewer = VideoPreviewer(self._ges_elem, self.timeline.app.settings.previewers_max_cpu)
         previewer.get_style_context().add_class("VideoUriSource")
 
         return previewer
@@ -782,9 +806,8 @@ class AudioUriSource(TimelineElement):
         self.get_style_context().add_class("AudioUriSource")
 
     def _getPreviewer(self):
-        previewer = AudioPreviewer(self._ges_elem)
+        previewer = AudioPreviewer(self._ges_elem, self.timeline.app.settings.previewers_max_cpu)
         previewer.get_style_context().add_class("AudioUriSource")
-        previewer.startLevelsDiscoveryWhenIdle()
 
         return previewer
 
@@ -908,11 +931,16 @@ class Clip(Gtk.EventBox, Zoomable, Loggable):
         if target.name() == EFFECT_TARGET_ENTRY.target:
             self.info("Adding effect %s", self.timeline.dropData)
             self.timeline.resetSelectionGroup()
+            self.timeline.current_group.add(self.ges_clip)
             self.timeline.selection.setSelection([self.ges_clip], SELECT)
             self.app.gui.switchContextTab(self.ges_clip)
 
-            self.app.gui.clipconfig.effect_expander.addEffectToClip(self.ges_clip,
-                                                                    self.timeline.dropData)
+            effect_info = self.app.effects.getInfo(self.timeline.dropData)
+            pipeline = self.timeline.ges_timeline.get_parent()
+            with self.app.action_log.started("add effect",
+                                             finalizing_action=CommitTimelineFinalizingAction(pipeline),
+                                             toplevel=True):
+                self.add_effect(effect_info)
             self.timeline.cleanDropData()
             success = True
 
@@ -920,6 +948,31 @@ class Clip(Gtk.EventBox, Zoomable, Loggable):
 
         return success
 
+    def add_effect(self, effect_info):
+        """Adds the specified effect if it can be applied to the clip.
+
+        Args:
+            effect_info (EffectInfo): The effect to add.
+        """
+        factory_name = effect_info.effect_name
+        if factory_name in ALLOWED_ONLY_ONCE_EFFECTS:
+            for effect in self.ges_clip.find_track_elements(None, GES.TrackType.VIDEO,
+                                                   GES.BaseEffect):
+                for elem in effect.get_nleobject().iterate_recurse():
+                    if elem.get_factory().get_name() == factory_name:
+                        self.error("Not adding %s as it would be duplicate"
+                                   " and this is not allowed.", factory_name)
+                        # TODO Let the user know about why it did not work.
+                        return effect
+
+        for track_element in self.ges_clip.get_children(False):
+            if effect_info.good_for_track_element(track_element):
+                # Actually add the effect
+                effect = GES.Effect.new(effect_info.bin_description)
+                self.ges_clip.add(effect)
+                return effect
+        return None
+
     def updatePosition(self):
         layer = self.layer
         if not layer or layer != self.get_parent():
